using System.Collections;
using System.Collections.Generic;
using Unity.Collections;
using UnityEngine;
using Unity.Mathematics;
using Unity.Jobs;
using UnityEditor;
using UnityEngine.Profiling;

public struct DODPathNode
{
    public bool canWalk;

    public int index;
    public int previousIndex;

    public float gCost;
    public float hCost;
    public float fCost;
    public float2 position;

    public void CalculateFCost()
    {
        fCost = gCost + hCost;
    }
}

public class SZM_PathFindingManager : CustomSingleton<SZM_PathFindingManager>
{
    private const float STRAIGHT_COST = 10;
    private const float DIAGONAL_COST = 14;

    protected override void Initialize()
    {
        #region DOD
#if UNITY_EDITOR
        RegisterEditorEvent();
#endif
        #endregion
    }

    #region DOD
    public NativeArray<DODPathNode> allDODNode;
    private int _mapMaxNodeLength;
    private float2 _bottomLeftPos;
    public void ConvertDataToDODNode(SZM_MapData mapData)
    {
        if (allDODNode.IsCreated)
        {
            allDODNode.Dispose();
        }

        int squareLength = SZM_TerrainMaker.squareLength;
        _mapMaxNodeLength = mapData.GetMapMaxLength() * squareLength;
        allDODNode = new NativeArray<DODPathNode>(_mapMaxNodeLength * _mapMaxNodeLength, Allocator.Persistent);

        float halfToNodeCenter = SZM_TerrainMaker.halfSquareLength - 0.5f;

        float minNodeCenterX = mapData.GetMinX() - halfToNodeCenter;
        float minNodeCenterY = mapData.GetMinY() - halfToNodeCenter;
        _bottomLeftPos = new Vector2(minNodeCenterX, minNodeCenterY);

        foreach (KeyValuePair<Vector2, SZM_TerrainDataCell> data in mapData.AllData)
        {
            for (int i = 0; i < data.Value.walkableData.Length; i++)
            {
                int row = i % squareLength;
                int column = i / squareLength;

                Vector2 nodeCenterPosition =
                    data.Key - new Vector2(halfToNodeCenter - column, 
                                           halfToNodeCenter - row);

                int index = GetDODIndex(nodeCenterPosition.x, nodeCenterPosition.y);

                DODPathNode tempNode = new DODPathNode();
                tempNode.canWalk = data.Value.walkableData[i] == 1;
                tempNode.gCost = float.MaxValue;
                tempNode.CalculateFCost();
                tempNode.previousIndex = -1;
                tempNode.position = nodeCenterPosition;
                tempNode.index = index;

                allDODNode[index] = tempNode;
            }
        }
    }

    private int GetDODIndex(float xPos, float yPos)
    {
        return (int)((xPos - _bottomLeftPos.x) * _mapMaxNodeLength + (yPos - _bottomLeftPos.y));
    }

    private float CalculateDODDistanceCost(float2 beginPos, float2 endPos)
    {
        float xDistance = math.abs(beginPos.x - endPos.x);
        float yDistance = math.abs(beginPos.y - endPos.y);
        float remaining = math.abs(xDistance - yDistance);
        return DIAGONAL_COST * math.min(xDistance, yDistance) + remaining * STRAIGHT_COST;
    }

    private int GetLowestFCostDODIndex(NativeList<int> openList)
    {
        DODPathNode lowestFCostNode = allDODNode[openList[0]];
        for (int i = 1; i < openList.Length; i++)
        {
            DODPathNode tempNode = allDODNode[openList[i]];
            if (lowestFCostNode.fCost > tempNode.fCost)
                lowestFCostNode = tempNode;
        }
       
        return lowestFCostNode.index;
    }
    
    public NativeList<float2> GetDODPath(float2 beginPos, float2 endPos)
    {
        Profiler.BeginSample("@@ 11");

        int startIndex = GetDODIndex(beginPos.x, beginPos.y);
        DODPathNode beginNode = allDODNode[startIndex];
        beginNode.gCost = 0;
        beginNode.hCost = CalculateDODDistanceCost(beginPos, endPos);
        beginNode.CalculateFCost();
        allDODNode[startIndex] = beginNode;

        NativeList<int> openList = new NativeList<int>(Allocator.Temp);
        NativeList<int> closeList = new NativeList<int>(Allocator.Temp);
        NativeArray<int2> nightborOffsetList = new NativeArray<int2>(new int2[]{
            new int2(1, 0),
            new int2(1, 1),
            new int2(0, 1),
            new int2(-1, 1),
            new int2(-1, 0),
            new int2(-1, -1),
            new int2(0, -1),
            new int2(1, -1),
        }, Allocator.Temp);

        openList.Add(beginNode.index);

        int endIndex = GetDODIndex(endPos.x, endPos.y);

        float lengthInterval = _mapMaxNodeLength - 1;
        float2 topRightPos =
            new float2(_bottomLeftPos.x + lengthInterval,
                       _bottomLeftPos.y + lengthInterval);

        Profiler.EndSample();


        Profiler.BeginSample("@@ 22");

        int count = 0;

        int nightborOffsetListCount = nightborOffsetList.Length;
        while (openList.Length > 0)
        {
            count++;

            int currentIndex = GetLowestFCostDODIndex(openList);
            DODPathNode currentNode = allDODNode[currentIndex];
            if (currentIndex == endIndex)
                break;

            for (int i = 0; i < openList.Length; i++)
            {
                if (openList[i] == currentIndex)
                {
                    openList.RemoveAtSwapBack(i);
                    break;
                }
            }
            closeList.Add(currentIndex);

            for (int i = 0; i < nightborOffsetListCount; i++)
            {
                float2 nighborPos = currentNode.position + nightborOffsetList[i];

                if (!InsideRange(nighborPos, topRightPos))
                    continue;

                int nightborIndex = GetDODIndex(nighborPos.x, nighborPos.y);

                if (closeList.Contains(nightborIndex))
                    continue;

                DODPathNode neighborNode = allDODNode[nightborIndex];
                if (!neighborNode.canWalk)
                    continue;

                float tempGCost = currentNode.gCost + CalculateDODDistanceCost(currentNode.position, nighborPos);
                if (tempGCost < neighborNode.gCost)
                {
                    neighborNode.previousIndex = currentIndex;
                    neighborNode.gCost = tempGCost;
                    neighborNode.hCost = CalculateDODDistanceCost(nighborPos, endPos);
                    neighborNode.CalculateFCost();
                    allDODNode[nightborIndex] = neighborNode;

                    if (!openList.Contains(nightborIndex))
                        openList.Add(nightborIndex);
                }
            }
        }
        Profiler.EndSample();

        Debug.LogError("Count = " + count);

        Profiler.BeginSample("@@ 33");
        openList.Dispose();
        closeList.Dispose();
        nightborOffsetList.Dispose();

        DODPathNode endNode = allDODNode[endIndex];
        Profiler.EndSample();

        return CalculateDODPath(endNode);
    }

    private NativeList<float2> CalculateDODPath(DODPathNode endNode)
    {
        if (endNode.previousIndex < 0)
        {
            return new NativeList<float2>(Allocator.Temp);
        }

        Profiler.BeginSample("@@ 44");

        NativeList<float2> retList = new NativeList<float2>(Allocator.Persistent);
        retList.Add(endNode.position);

        DODPathNode currentNode = endNode;
        while (currentNode.previousIndex >= 0)
        {
            currentNode = allDODNode[currentNode.previousIndex];
            retList.Add(currentNode.position);
        }
        Profiler.EndSample();

        return retList;
    }


    private bool InsideRange(float2 position, float2 topRightPos)
    {
        return !(position.x < _bottomLeftPos.x ||
                 position.y < _bottomLeftPos.y ||
                 position.x > topRightPos.x ||
                 position.y > topRightPos.y);
    }

    private void DisposeAllNode()
    {
        if (allDODNode.IsCreated)
        {
            allDODNode.Dispose();
        }
    }
#if UNITY_EDITOR
    private void RegisterEditorEvent()
    {
        EditorApplication.playModeStateChanged += ModeChanged;
    }

    private void ModeChanged(PlayModeStateChange playModeState)
    {
        if (playModeState == PlayModeStateChange.EnteredEditMode)
        {
            DisposeAllNode();
        }
    }
#endif

    #endregion



    #region OOP
    public Dictionary<Vector2, SZM_PathNode> AllNodes { get; private set; } = new Dictionary<Vector2, SZM_PathNode>();

    public void ConvertDataToNode(SZM_MapData mapData)
    {
        AllNodes.Clear();

        int squareLength = SZM_TerrainMaker.squareLength;

        foreach (KeyValuePair<Vector2, SZM_TerrainDataCell> data in mapData.AllData)
        {
            for (int i = 0; i < data.Value.walkableData.Length; i++)
            {
                int row = i % squareLength;
                int column = i / squareLength;

                Vector2 bottomLeftPosition = 
                    data.Key - new Vector2(SZM_TerrainMaker.halfSquareLength, SZM_TerrainMaker.halfSquareLength);

                Vector2 nodeCenterPosition =
                    new Vector2(bottomLeftPosition.x + column + 0.5f,
                                bottomLeftPosition.y + row + 0.5f);

                SZM_PathNode tempNode = new SZM_PathNode();
                tempNode.canWalk = data.Value.walkableData[i] == 1;
                tempNode.gCost = float.MaxValue;
                tempNode.CalculateFCost();
                tempNode.position = nodeCenterPosition;

                AllNodes.Add(nodeCenterPosition, tempNode);
            }
        }
    }

    public SZM_PathNode GetNode(Vector2 postion)
    {
        AllNodes.TryGetValue(postion, out SZM_PathNode retNode);
        return retNode;
    }

    public Stack<SZM_PathNode> GetPath(SZM_PathNode beginNode, SZM_PathNode endNode)
    {
        if (!beginNode.canWalk)
        {
            Debug.LogError("Begin Node not walkable!");
            return null;
        }

        Dictionary<Vector2, SZM_PathNode> openDic = new Dictionary<Vector2, SZM_PathNode>();
        Dictionary<Vector2, SZM_PathNode> closeDic = new Dictionary<Vector2, SZM_PathNode>();
        openDic.Add(beginNode.position, beginNode);

        beginNode.gCost = 0;
        beginNode.hCost = CalculateDistanceCost(beginNode, endNode);
        beginNode.CalculateFCost();

        int count = 0;

        while (openDic.Count > 0)
        {
            count++;
            SZM_PathNode currentNode = GetLowestFCostPosition(openDic);
            if (currentNode.position == endNode.position)
            {
                Debug.LogError("Count = " + count);
                return CalculatePath(endNode);
            }

            closeDic.Add(currentNode.position, openDic[currentNode.position]);
            openDic.Remove(currentNode.position);

            neighborNodeList.Clear();
            // Left, right, down, up
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Left));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Right));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Up));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Down));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Right_Up));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Right_Down));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Left_Up));
            TryAddNodeToList( neighborNodeList, currentNode.position.GetDirection(Direction.Left_Dowm));

            for (int i = 0; i < neighborNodeList.Count; i++)
            {
                if (closeDic.ContainsKey(neighborNodeList[i].position)) continue;

                float tempGCost = currentNode.gCost + CalculateDistanceCost(currentNode, neighborNodeList[i]);

                if (tempGCost < neighborNodeList[i].gCost)
                {
                    neighborNodeList[i].previousNode = currentNode;
                    neighborNodeList[i].gCost = tempGCost;
                    neighborNodeList[i].hCost = CalculateDistanceCost(neighborNodeList[i], endNode);
                    neighborNodeList[i].CalculateFCost();

                    openDic.TryAdd(neighborNodeList[i].position, neighborNodeList[i]);
                }
            }
        }

        return null;
    }

    private List<SZM_PathNode> neighborNodeList = new List<SZM_PathNode>(8);
    public void TryAddNodeToList(List<SZM_PathNode> targetList, Vector2 keyPosition)
    {
        if (AllNodes.TryGetValue(keyPosition, out SZM_PathNode node) &&
           node.canWalk)
        {
            targetList.Add(node);
        }
    }

    private Stack<SZM_PathNode> CalculatePath(SZM_PathNode endNode)
    {
        Stack<SZM_PathNode> path = new Stack<SZM_PathNode>();
        path.Push(endNode);
        SZM_PathNode currentNode = endNode;
        while (currentNode.previousNode != null)
        {
            path.Push(currentNode.previousNode);
            currentNode = currentNode.previousNode;
        }
        return path;
    }

    private SZM_PathNode GetLowestFCostPosition(Dictionary<Vector2, SZM_PathNode> dic)
    {
        SZM_PathNode retNode = null;
        float lowestFCost = float.MaxValue;
        foreach (KeyValuePair<Vector2, SZM_PathNode> data in dic)
        {
            if (data.Value.fCost < lowestFCost)
            {
                lowestFCost = data.Value.fCost;
                retNode = data.Value;
            }
        }
        return retNode;
    }

    private float CalculateDistanceCost(SZM_PathNode beginNode, SZM_PathNode endNode)
    {
        float xDistance = Mathf.Abs(beginNode.position.x - endNode.position.x);
        float yDistance = Mathf.Abs(beginNode.position.y - endNode.position.y);
        float remaining = Mathf.Abs(xDistance - yDistance);
        return DIAGONAL_COST * Mathf.Min(xDistance, yDistance) + remaining * STRAIGHT_COST;
    }

    #endregion
}

public static class SZM_PathFIndingExtension
{
    public static Vector2 GetDirection(this Vector2 centerPositin, Direction direction)
    {
        return new Vector2(centerPositin.x + ((direction & Direction.Right) > 0 ? 1 : 0) - ((direction & Direction.Left) > 0 ? 1 : 0),
                           centerPositin.y + ((direction & Direction.Up) > 0 ? 1 : 0) - ((direction & Direction.Down) > 0 ? 1 : 0));
    }

    public static SZM_PathNode GetNode(this Vector2 postion)
    {
        return SZM_PathFindingManager.Instance.GetNode(postion);
    }
}


public enum Direction
{
    Up          = 0x00000001,
    Right       = 0x00000002,
    Down        = 0x00000004,
    Left        = 0x00000008,
    Right_Up    = 0x00000003,
    Right_Down  = 0x00000006,
    Left_Dowm   = 0x0000000c,
    Left_Up     = 0x00000009,
}